/*-*-c++-*-*/
/******************************************************************************
 **                                                                          **
 ** Copyright (c) 2000 Massachusetts Institute of Technology, Cambridge, MA. **
 ** All rights reserved.                                                     **
 **                                                                          **
 ******************************************************************************
  ==========================================================================
  ============================= File Info ==================================

  Author: Zhenhai Zhu
  
  Description:

  interface for superLU c code

  Resources:

  See also:

  const static char cvsid[] = "$Id$";

  ==========================================================================
*/

#ifndef _SUPER_LU_H_
#define _SUPER_LU_H_

#include <complex>
// I could not include both files because they share the same variable names.
// Otherwise this single header file would be able to handle both double and complex.
#include "zsp_defs.h" // interface to superLU
#include "dsp_defs.h" // interface to superLU 
#include "util.h" // interface to superLU
#include "utils.h" // this is for pfft::errorMessage
#include "spColMat.h"

namespace superLU {

  template <class T>
  class SuperLU {

    typedef T value_type;

  public:
    SuperLU(void) : memoryAllocated(false) {}
    ~SuperLU(void) {
      if (memoryAllocated) {
	Destroy_CompCol_Matrix(&A);
	Destroy_SuperNode_Matrix(&L);
	Destroy_CompCol_Matrix(&U);
    
	//    Destroy_Dense_Matrix(&X); 
	// this is redundant because its store has been freed in solve(), 
	// and the nzval will be freed since it is the same as superLUrhs
	free(dSuperLUrhs); 
	free(zSuperLUrhs); 
	free(perm_c);
	free(perm_r);
	free(etree);
      }
    }

    /**********************************************************************
     * setup --
     * It apears that superLU only support LU of compressed column format.
     **********************************************************************/
    //
    // specialization for complex
    //
    void 
    setup(
	  pfft::SpColMat<std::complex<double> >& preCondMat) 
    {
      numRow = preCondMat.numRow();
      numCol = preCondMat.numCol();
      numNonZero = preCondMat.numNonZero();

      if (!memoryAllocated) allocate(typeFlag_);

      int elementCount = 0;
      for (size_t col = 0; col < numCol; col++) {
	colStart[col] = elementCount;
	for (size_t row = 0; row < preCondMat[col].size(); row++) {
	  zmat[elementCount].r = preCondMat[col].value(row).real();
	  zmat[elementCount].i = preCondMat[col].value(row).imag();
	  colIndex[elementCount] = preCondMat[col].index(row);
	  elementCount++;
	}
      }
      colStart[preCondMat.numCol()] = elementCount;

      // Structure A has its internal pointers pointing to mat, colIndex and 
      // colStart, so there is no redundant memory allocation. 
      // This also means that I should only deallocate A. 
      // Otherwie I will free dandling pointer. This will cause segmentation fault.
      zCreate_CompCol_Matrix(&A, numRow, numCol, numNonZero, zmat, colIndex, 
			     colStart, NC, _Z, GE);

      factor(typeFlag_);
    }

    //
    // specialization for double
    //
    void 
    setup (
	   pfft::SpColMat<double>& preCondMat)
    {
      numRow = preCondMat.numRow();
      numCol = preCondMat.numCol();
      numNonZero = preCondMat.numNonZero();

      if (!memoryAllocated) allocate(typeFlag_);

      int elementCount = 0;
      for (size_t col = 0; col < numCol; col++) {
	colStart[col] = elementCount;
	for (size_t row = 0; row < preCondMat[col].size(); row++) {
	  dmat[elementCount] = preCondMat[col].value(row);
	  colIndex[elementCount] = preCondMat[col].index(row);
	  elementCount++;
	}
      }
      colStart[preCondMat.numCol()] = elementCount;

      // Structure A has its internal pointers pointing to mat, colIndex and 
      // colStart, so there is no redundant memory allocation. 
      // This also means that I should only deallocate A. 
      // Otherwie I will free dandling pointer. This will cause segmentation fault.
      dCreate_CompCol_Matrix(&A, numRow, numCol, numNonZero, dmat, colIndex, 
			     colStart, NC, _D, GE);

      factor(typeFlag_);
    }

    /**********************************************************************
     * solve --
     **********************************************************************/
    //
    // specialization for complex
    //
    void
    solve (
	   const TNT::Vector<std::complex<double> >& rhs,
	   TNT::Vector<std::complex<double> >& sol)
    {
      for (size_t i = 0; i < numRow; i++) {
	zSuperLUrhs[i].r = rhs[i].real();
	zSuperLUrhs[i].i = rhs[i].imag();
      }
      zCreate_Dense_Matrix(&X, numRow, 1, zSuperLUrhs, numRow, DN, _Z, GE);
    
      int info;
      char trant[1];
      *trant = 'N';
      zgstrs (trant, &L, &U, perm_r, perm_c, &X, &info);
      if (info) {
	pfft::errorMessage("SuperLU::solve()",
			   "SuperLU fails to solve the sparse matrix equation");
      }

      DNformat *Xstore = (DNformat *) X.Store;
      doublecomplex* zSuperLUsol = (doublecomplex *) Xstore->nzval;
      for (size_t i = 0; i < numRow; i++) {
	sol[i] = std::complex<double>(zSuperLUsol[i].r, zSuperLUsol[i].i);
      }

      Destroy_SuperMatrix_Store(&X);
    }

    //
    // specialization for double
    //
    void
    solve (
	   const TNT::Vector<double>& rhs,
	   TNT::Vector<double>& sol)
    {
      for (size_t i = 0; i < numRow; i++) {
	dSuperLUrhs[i] = rhs[i];
      }
      dCreate_Dense_Matrix(&X, numRow, 1, dSuperLUrhs, numRow, DN, _D, GE);
    
      int info;
      char trant[1];
      *trant = 'N';
      dgstrs (trant, &L, &U, perm_r, perm_c, &X, &info);
      if (info) {
	pfft::errorMessage("SuperLU::solve()",
			   "SuperLU fails to solve the sparse matrix equation");
      }

      DNformat *Xstore = (DNformat *) X.Store;
      double* superLUsol = (double *) Xstore->nzval;
      for (size_t i = 0; i < numRow; i++) {
	sol[i] = superLUsol[i];
      }

      Destroy_SuperMatrix_Store(&X);
    }

    double estimateCondNum_OneNorm(void);
    double estimateCondNum_InfNorm(void);

  private:
    bool memoryAllocated;
    int numRow;
    int numCol;
    int numNonZero;
    SuperMatrix A, L, U, X;
    int *colIndex, *colStart;
    int *perm_c, *perm_r, *etree;
    doublecomplex *zmat;
    double *dmat;
    doublecomplex *zSuperLUrhs, *zSuperLUsol;
    double *dSuperLUrhs, *dSuperLUsol;
    T typeFlag_;

    /**********************************************************************
     * allocate --
     **********************************************************************/
    //
    // specialization for double
    //
    void allocate (double flag) {
      allocateSharedPart();
      dmat = (double *) malloc(numNonZero * sizeof(double));
      dSuperLUrhs = (double*) malloc(numRow * sizeof(double));
      memoryAllocated = true;  
    }

    //
    // specialization for complex
    //
    void allocate (std::complex<double> flag) {
      allocateSharedPart();
      zmat = (doublecomplex *) malloc(numNonZero * sizeof(doublecomplex));
      zSuperLUrhs = (doublecomplex*)malloc(numRow * sizeof(doublecomplex));
      memoryAllocated = true;  
    }

    void allocateSharedPart (void) {
      colIndex = (int *) malloc(numNonZero * sizeof(int));
      colStart = (int *) malloc((numCol+1) * sizeof(int));
      perm_r = (int*) malloc(numRow * sizeof(int)); 
      perm_c = (int*) malloc(numCol * sizeof(int)); 
      etree = (int*) malloc(numCol * sizeof(int));
    }

  };

} // namespace superLU

#endif
