/*-*-c++-*-*/
/******************************************************************************
 **                                                                          **
 ** Copyright (c) 2000 Massachusetts Institute of Technology, Cambridge, MA. **
 ** All rights reserved.                                                     **
 **                                                                          **
 ******************************************************************************
  ==========================================================================
  ============================= File Info ==================================

  Author: Zhenhai Zhu
  
  Description:

  Resources:

  See also:

  ==========================================================================
*/

const static char cvsid[] = "$Id$";

#include "main.h"

using namespace std;
using namespace pfft;
using namespace TNT;

static char* elementFileName = 0;

/**********************************************************************
 * main --
 **********************************************************************/
int
main(
     int argc,
     char *argv[])
{  
  cmdLineParsor(argc, argv);
  
  // read discretization from a file
  vector<element> srcElementList;
  string title;
  size_t numCond;
  try {
    readElementFile(elementFileName, srcElementList, title, numCond);
  } catch (const std::domain_error error) {
    cerr << error.what() << endl;
    cout << "Wrong command line option! Illegal input file name" << endl
	 << "exit ..." << endl;
    exit(1);
  }

  // calculate capacitance and output
  TNT::Matrix<double> capMatrix(numCond, numCond, 0.);
  compCap(numCond, srcElementList, capMatrix);
  resultsReport(capMatrix, title);
  
  return 0;
}

/**********************************************************************
 * compCap --
 **********************************************************************/
void 
compCap (
	 const size_t numCond,
	 std::vector<pfft::element>& srcElementList,
	 TNT::Matrix<double>& capMatrix)
{
  // for capacitance problem, evaluation panels are the same as source panels
  std::vector<pfft::element>& evalElementList = srcElementList;

  // Three steps are involved in defining what problem we are trying to solve. 
  // Replace them with your own 

  // step 1: define kernel
  // The green's function is 1/r. And both inner and outer operators are just 
  // identity, or NONE as called in pfft++
  pfft::OneOverR kernel;
  std::vector<pfft::DifferentialOperator> innerOperatorList;
  std::vector<pfft::DifferentialOperator> outerOperatorList;
  outerOperatorList.push_back(pfft::NONE);
  innerOperatorList.push_back(pfft::NONE);

  // step 2: define panel integration
  // pfft++ provides a piece wise collocation panel integration routine for 1/r kernel
  typedef pfft::StaticCollocation<pfft::element> Calcp;
  Calcp calcp(outerOperatorList, innerOperatorList);

  // step 3: define pfft 
  // Use the default accuracy level for matrix vector product.
  // For 1/r kernel, the accuracy should be about 4-5 digit.
  typedef pfft::Pfft<double, double, pfft::OneOverR, Calcp> PfftType;
  PfftType pfft(srcElementList, evalElementList, kernel, calcp);

  // define the interface between pfft and gmres
  FastMV<PfftType> fastMatVecProd(pfft);

  // This defines the preconditioner. It is again problem-dependent.
  PreconditionerType precondType = RIGHT_PRE_CONDITIONER;
  Preconditioner<Calcp> precond(srcElementList, evalElementList, calcp, precondType);

  TNT::Vector<double> RHS(srcElementList.size());
#ifdef COND_INDEX_OFFSET_1
  size_t startIndex = 1;
  size_t endIndex = numCond;
#else
  size_t startIndex = 0;
  size_t endIndex = numCond - 1;
#endif
  for(size_t condIndex = startIndex; condIndex <= endIndex; condIndex++) {
    // set voltage on one specified conductor as 1 volt
    setRHS(RHS, condIndex, evalElementList);  
  
    // restart after every 30 iterations so we don't have to use a lot of memory 
    // to store Hessenberg matrix.
    size_t restart = 30; 
    size_t max_iter = 10*restart; // quit after 10 restart.
    double tol = 1e-3; // this means 3-digit accuracy in gmres iterations
    // initialize solution to be zero. It is the initial guess in Gmres.
    TNT::Vector<double> solution(evalElementList.size(), 0.);
    gmres(fastMatVecProd, solution, RHS, precond, max_iter, restart, tol);

    getOneColumnOfCapMatrix(capMatrix, condIndex, solution, srcElementList);
  }
}

/**********************************************************************
 * setRHS --
 * Fill the right hand vector
 * In capacitance problem, the given is the potential on the specified
 * conductor surface, whose index is sCondIndex. Usually the potential
 * value is just 1 so that the calculated charge is one column of the 
 * capacitance matrix.
 **********************************************************************/
void 
setRHS (
	TNT::Vector<double>& RHS,
	const size_t sCondIndex,
	const std::vector<pfft::element>& evalElementList)
{
  for (size_t i=0; i < evalElementList.size(); i++) {
    if(evalElementList[i].boundaryIndex() == sCondIndex) {
      RHS[i] = 1.;
    } else {
      RHS[i] = 0.;
    }
  }
}

/**********************************************************************
 * getOneColumnOfCapMatrix --
 **********************************************************************/
void 
getOneColumnOfCapMatrix(
			TNT::Matrix<double>& capMatrix,
			const size_t sCondIndex,
			const TNT::Vector<double>& solution,
			const std::vector<pfft::element>& srcElementList)
{
  // note: capMatrix must be set to zero before calculation.
  double coef = 4. * 3.1415926 * 8.854; // 4*PI*epsilon_0
  TNT::Vector<double>::const_iterator iterx = solution.begin();
  std::vector<pfft::element>::const_iterator itery = srcElementList.begin();
  for(; iterx != solution.end(); ++iterx, ++itery) {
    size_t condIndex = itery->boundaryIndex();
    capMatrix[condIndex-1][sCondIndex-1] += (*iterx) * itery->area() * coef;
  }
}

/**********************************************************************
 * resultsReport --
 **********************************************************************/
void 
resultsReport (
	       const TNT::Matrix<double>& capMat,
	       const std::string& title)
{
  ofstream fout("cap.dat",ios::app);
  fout << endl << endl << elementFileName << ": " << title << endl << endl;
  cout << endl << endl << elementFileName << ": " << title << endl << endl;
  for(size_t ii=0; ii<capMat.num_rows(); ++ii) {
    for(size_t jj=0; jj<capMat.num_cols(); ++jj) {
      fout<<"C["<<ii<<"]["<<jj<<"] = "<<capMat[ii][jj]<<"  ";
      cout<<"C["<<ii<<"]["<<jj<<"] = "<<capMat[ii][jj]<<"  ";
    }
    fout<<"\n\n";
    cout<<"\n\n";
  }
  fout<<endl;
  cout<<endl;
  fout.close();

  cout<< "\tResults are all in pF" << endl;
  cout<< "\tResults are also appended to the file cap.dat" << endl;
}

/**********************************************************************
 * cmdLineParsor --
 **********************************************************************/
void
cmdLineParsor (
	       int argc,
	       char *argv[])
{
  int option;

  if (argc == 1) {
    printUsage();	
    exit(0);
  }	

  opterr = 0;  /* Disable stderr Error Message by getopt */
  while ((option = getopt(argc, argv, "i:h")) != EOF) {
    switch (option) {
    case INPUT_FILE_NAME:
      elementFileName = optarg;
      break;
    case HELP:
      printUsage();
      exit(0);
      break;
    case '?': /* unknown option */
      printf("\n\n\t Warning in pfft:"
	     "\n\t   Wrong command line options"
	     "\n\t   please check.\n");
      printUsage();
      exit(1);
      break;
    default: /* should never happen */
      printUsage();
      exit(2);
    }
  }

}

/**********************************************************************
 * printUsage --
 **********************************************************************/
void
printUsage (
	    void)
{
  printf("\n\tUsage:  fftcap++ <options> \n"
	 "\tWhere options are:\n"
	 "\t -i <input file name>\n"
	 "\t -h Print this message and Exit\n"
	 );
}




