/*-*-c++-*-*/
/******************************************************************************
 **                                                                          **
 ** Copyright (c) 2000 Massachusetts Institute of Technology, Cambridge, MA. **
 ** All rights reserved.                                                     **
 **                                                                          **
 ******************************************************************************
  ==========================================================================
  ============================= File Info ==================================

  Author: Zhenhai Zhu
  
  Description:

  A good preconditioner needs at least two crucial ingredients: 
  1) extracting essential information from the system matrix and store it in 
     a sparse data structure
  2) A very efficient sparse LU solver

  In this simple preconditioner class, I only extract the diagonal of the system
  matrix. It turns iut this is a pretty good one.
  In theory, no LU factorization is needed for a diagonal matrix. But to show 
  how to use the interface class SuperLU, I delibrately treat this diag matrix
  as if it is a non-diag sparse matrix.

  If you use this class as a template, you just need to worry about what to put 
  into the sparse matrix precondMat_.

  Resources:

  See also:

  const static char cvsid[] = "$Id:$";

  ==========================================================================
*/

#ifndef _PRECOND_H_
#define _PRECOND_H_

#include "utils.h" 
#include "spColMat.h"
#include "realsuperLU.h"

enum PreconditionerType { NO_PRE_CONDITIONER, LEFT_PRE_CONDITIONER,
			  RIGHT_PRE_CONDITIONER, NUM_PRE_CONDITIONER_TYPE };

template <class Integration>
class Preconditioner { 

public:
  Preconditioner(
		     std::vector<pfft::element>& srcElementList,
		     std::vector<pfft::element>& evalElementList,
		     Integration& calcp,
		     const PreconditionerType& preCondType)
    : precondType_(preCondType), calcp_(calcp) 
  {
    // fill a sparse matrix and prepare to feed it to SuperLU
    size_t integralIndex = calcp_.findIntegralIndex(pfft::NONE, pfft::NONE);
    size_t numHostElement = srcElementList.size();
    precondMat_ = pfft::SpColMat<double>(numHostElement, numHostElement);
    for (size_t hostIndex = 0; hostIndex < numHostElement; hostIndex++) {
      const pfft::element& srcElement = srcElementList[hostIndex];
      calcp_(srcElement, srcElement);
      size_t rowIndex = hostIndex;
      size_t colIndex = rowIndex;
      double diagElement = calcp_.result(integralIndex);
      precondMat_.insertElement(rowIndex, colIndex, diagElement);
    }

    // set up superLU class, copy the matrix to superLU's data structure and 
    // do LU factorization for just once.
    TNT::stopwatch time;  time.start();
    superLU_.setup(precondMat_);
    superLU_.factor();
    pfft::timeReport("Time used for LU factorizing the pre-conditioner matrix := ", 
		     time.read());
  };

  const bool isLeftPreconditioner(void) const {return precondType_ == LEFT_PRE_CONDITIONER; }
  const bool isRightPreconditioner(void) const {return precondType_ == RIGHT_PRE_CONDITIONER; }
  
  template<class T> 
  void solve(
	     const TNT::Vector<T>& rhs,
	     TNT::Vector<T>& sol)
  {
    if (precondType_ != NO_PRE_CONDITIONER) {
      superLU_.solve(rhs, sol);      
    } else {
      if (&sol != &rhs) {
	for (size_t i = 0; i < rhs.size(); i++) {
	  sol[i] = rhs[i];
	}
      }	
    }
  }
  
private:
  PreconditionerType precondType_;
  Integration& calcp_;
  pfft::SpColMat<double> precondMat_;
  superLU::SuperLU superLU_;
};

#endif
